/* File:                $Id: IndexServer.java 2502 2012-09-25 15:57:46Z svc $
 * Revision:            $Revision: 2502 $
 * Author:              $Author: svc $
 * Date:                $Date: 2012-09-25 17:57:46 +0200 (Tue, 25 Sep 2012) $
 *
 * The Netarchive Suite - Software to harvest and preserve websites
 * Copyright 2004-2012 The Royal Danish Library, the Danish State and
 * University Library, the National Library of France and the Austrian
 * National Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 
 *  USA
 */
package dk.netarkivet.harvester.indexserver;

import dk.netarkivet.common.distribute.indexserver.RequestType;
import dk.netarkivet.common.utils.CleanupIF;

import java.util.HashSet;
import java.util.Set;

/** Index server.
 * Handles request for lucene indexes of crawl logs and cdx indexes of jobs,
 * using two multifilebasedcache objects as handlers.
 *
 * The server sets up handlers for three kinds of indexes (as defined by
 * RequestType):
 * A CDX index, where each index is one file, gzip-compressed.
 * A DEDUP_CRAWL_LOG index, where each index is multiple files, gzip-compressed,
 * making up a Lucene index of non-HTML files.
 * A FULL_CRAWL_LOG index, where each index is multiple files, gzip-compressed,
 * making up a Lucene index of all files.
 * */
public class IndexServer implements CleanupIF{
    /** The remote server that hands us indexes. */
    private IndexRequestServerInterface remoteServer;
    /** The singleton instance of this class. */
    private static IndexServer instance;

    /** Instantiates the two handlers, and starts listening for requests. */
    protected IndexServer() {
        FileBasedCache<Set<Long>> cdxCache = new CDXIndexCache();
        FileBasedCache<Set<Long>> dedupCrawlLogCache
                = new DedupCrawlLogIndexCache();
        FileBasedCache<Set<Long>> fullCrawlLogCache
                = new FullCrawlLogIndexCache();
        //prompt the empty indices to pre-generated
        Set<Long> emptySet = new HashSet<Long>();
        cdxCache.getIndex(emptySet);
        dedupCrawlLogCache.getIndex(emptySet);
        fullCrawlLogCache.getIndex(emptySet);
                 
        remoteServer = IndexRequestServerFactory.getInstance();

        remoteServer.setHandler(RequestType.CDX, cdxCache);
        remoteServer.setHandler(RequestType.DEDUP_CRAWL_LOG,
                dedupCrawlLogCache);
        remoteServer.setHandler(RequestType.FULL_CRAWL_LOG,
                fullCrawlLogCache);
        remoteServer.start();
    }

    /** Get the unique index server instance.
     *
     * @return The instance;
     */
    public static synchronized IndexServer getInstance() {
        if (instance == null) {
            instance = new IndexServer();
        }
        return instance;
    }

    /** Close the server.
     */
    public void cleanup() {
        remoteServer.close();
        instance = null;
    }
}
