/* File:        $Id: NetarchiveSuiteUrlOperations.java 2251 2012-02-08 13:03:03Z mss $
 * Revision:    $Revision: 2251 $
 * Author:      $Author: mss $
 * Date:        $Date: 2012-02-08 14:03:03 +0100 (Wed, 08 Feb 2012) $
 *
 * Copyright 2004-2012 The Royal Danish Library, the Danish State and
 * University Library, the National Library of France and the Austrian
 * National Library.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package dk.netarkivet.wayback.batch.copycode;

import java.util.regex.Pattern;
import java.util.regex.Matcher;

import org.archive.net.UURI;
import org.apache.commons.httpclient.URIException;

/**
 * This is a cut'n'paste job from the wayback subversion repository, with calls
 * to UURIFactory replaced with calls to NetarchiveSuiteUURIFactory. Calls to
 * it can be replaced with calls to
 * org.archive.wayback.util.url.UrlOperations when we have an execution
 * environment which allows the reading of system properties
 * 
 */

public class NetarchiveSuiteUrlOperations {

    public static final String DNS_SCHEME = "dns:";

    public static final String HTTP_SCHEME = "http://";

    public static final String HTTPS_SCHEME = "https://";

    public static final String FTP_SCHEME = "ftp://";

    public static final String MMS_SCHEME = "mms://";

    public static final String RTSP_SCHEME = "rtsp://";

    // go brewster
    public static final String WAIS_SCHEME = "wais://";

    public static final String ALL_SCHEMES[] = { HTTP_SCHEME, HTTPS_SCHEME,
            FTP_SCHEME, MMS_SCHEME, RTSP_SCHEME, WAIS_SCHEME };


    public static final char PORT_SEPARATOR = ':';

    public static final char PATH_START = '/';


    private static final String CC_TLDS = "ac|ad|ae|af|ag|ai|al|am|an|ao|aq"
            + "|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs"
            + "|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cu|cv|cx"
            + "|cy|cz|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo"
            + "|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk"
            + "|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg"
            + "|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma"
            + "|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz"
            + "|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm"
            + "|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj"
            + "|sk|sl|sm|sn|so|sr|st|su|sv|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn"
            + "|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|um|us|uy|uz|va|vc|ve|vg|vi|vn|vu"
            + "|wf|ws|ye|yt|yu|za|zm|zw";

    private static final String GEN_TLDS = "aero|biz|cat|com|coop|edu|gov" 
        + "|info|int|jobs|mil|mobi|museum|name|net|org|pro|travel";


    private static final String ALL_TLD_PATTERN = CC_TLDS + "|" + GEN_TLDS;

    private static final String IP_PATTERN
        = "[0-9]+\\.[0-9]+\\.[0-9]+\\.[0-9]+";

    private static final Pattern AUTHORITY_REGEX =
        Pattern.compile("(([0-9a-z_.-]+)\\.(" + ALL_TLD_PATTERN + "))|" 
                + "(" + IP_PATTERN + ")");

    /**
     * @param urlPart
     * @return boolean indicating whether urlPart might be an Authority.
     */
    public static boolean isAuthority(String urlPart) {
        Matcher m = AUTHORITY_REGEX.matcher(urlPart);

        return (m != null) && m.matches();
    }

    /**
     * @param baseUrl
     * @param url
     * @return url resolved against baseUrl, unless it is absolute already
     */
    public static String resolveUrl(String baseUrl, String url) {
        for (final String scheme : ALL_SCHEMES) {
            if (url.startsWith(scheme)) {
                try {
                    return NetarchiveSuiteUURIFactory.getInstance(url)
                            .getEscapedURI();
                } catch (URIException e) {
                    e.printStackTrace();
                    // can't let a space exist... send back close to whatever
                    // came
                    // in...
                    return url.replace(" ", "%20");
                }
            }
        }
        UURI absBaseURI;
        UURI resolvedURI = null;
        try {
            absBaseURI = NetarchiveSuiteUURIFactory.getInstance(baseUrl);
            resolvedURI = NetarchiveSuiteUURIFactory.getInstance(absBaseURI,
                    url);
        } catch (URIException e) {
            e.printStackTrace();
            return url.replace(" ", "%20");
        }
        return resolvedURI.getEscapedURI();
    }

    public static String urlToScheme(final String url) {
        for (final String scheme : ALL_SCHEMES) {
            if (url.startsWith(scheme)) {
                return scheme;
            }
        }
        return null;
    }

    public static int schemeToDefaultPort(final String scheme) {
        if (scheme.equals(HTTP_SCHEME)) {
            return 80;
        }
        if (scheme.equals(HTTPS_SCHEME)) {
            return 443;
        }
        if (scheme.equals(FTP_SCHEME)) {
            return 21;
        }
        if (scheme.equals(RTSP_SCHEME)) {
            return 554;
        }
        if (scheme.equals(MMS_SCHEME)) {
            return 1755;
        }
        return -1;
    }

    public static String urlToHost(String url) {
        if (url.startsWith("dns:")) {
            return url.substring(4);
        }
        for (String scheme : ALL_SCHEMES) {
            if (url.startsWith(scheme)) {
                int hostIdx = scheme.length();
                int portIdx = url.indexOf(PORT_SEPARATOR, hostIdx + 1);
                int pathIdx = url.indexOf(PATH_START, hostIdx + 1);
                if (portIdx == -1 && pathIdx == -1) {
                    return url.substring(hostIdx);
                }
                if (portIdx == -1) {
                    return url.substring(hostIdx, pathIdx);
                }
                if (pathIdx == -1) {
                    return url.substring(hostIdx, portIdx);
                }
                if (pathIdx > portIdx) {
                    return url.substring(hostIdx, portIdx);
                } else {
                    return url.substring(hostIdx, pathIdx);
                }
            }
        }
        return url;
    }
}
